use bevy_eventwork::{tcp::TcpProvider, Network};
use std::{
	io::{stdin, stdout},
	process,
	sync::{mpsc, Arc, Mutex},
	thread,
};
use termion::{clear, cursor, event::Key, input::TermRead, raw::IntoRawMode};

use bevy::prelude::*;

use crate::common::client_msg::Move;

use super::net::Server;

pub struct InputPlugin;

impl Plugin for InputPlugin {
	fn build(&self, app: &mut App) {
		let rec = spawn_reader();
		let rec = Arc::new(Mutex::new(rec));
		app.insert_resource(InputReceiver(rec))
			.add_event::<KeyEvent>()
			.add_systems(Update, try_read_keys)
			.add_systems(Update, on_move);
	}
}

#[derive(Resource)]
pub struct InputReceiver(Arc<Mutex<mpsc::Receiver<Key>>>);

#[derive(Debug, Event)]
pub struct KeyEvent(Key);

fn spawn_reader() -> mpsc::Receiver<Key> {
	let (tx, rx) = mpsc::channel();

	let (exit_tx, exit_rx) = mpsc::channel();
	thread::spawn(move || {
		let raw_mode = stdout().into_raw_mode().unwrap();
		println!("{}", cursor::Hide);
		exit_rx.recv().ok();
		drop(raw_mode);
		println!("{}{}", cursor::Restore, clear::All);
		process::exit(0);
	});

	let exit_tx_ = exit_tx.clone();
	ctrlc::set_handler(move || {
		exit_tx_.send(()).ok();
	})
	.unwrap();

	thread::spawn(move || {
		let mut keys = stdin().keys();
		while let Some(Ok(key)) = keys.next() {
			if key == Key::Esc {
				exit_tx.send(()).ok();
			}
			tx.send(key).ok();
		}
	});

	rx
}

fn try_read_keys(receiver: Res<InputReceiver>, mut writer: EventWriter<KeyEvent>) {
	let receiver = receiver.0.lock().expect("Poisoned threads doomed the state.");
	writer.send_batch(receiver.try_iter().map(KeyEvent));
}

fn on_move(mut keys: EventReader<KeyEvent>, server: Res<Server>, net: Res<Network<TcpProvider>>) {
	if let Some(server) = server.0 {
		for key in keys.read() {
			let dir = match key.0 {
				Key::Char('z') => IVec2::Y,
				Key::Char('s') => -IVec2::Y,
				Key::Char('q') => -IVec2::X,
				Key::Char('d') => IVec2::X,
				_ => continue,
			};
			net.send_message(server, Move(dir)).unwrap();
		}
	}
}
