use bevy::{prelude::*, utils::HashSet};

use super::{
	display::RenderUpdate,
	map::{self, spawn_block},
	physics::Pos,
};

pub struct GenerationPlugin;

impl Plugin for GenerationPlugin {
	fn build(&self, app: &mut bevy::prelude::App) {
		app.insert_resource(GeneratedRegistry::default())
			.add_systems(Update, chunk_loading);
	}
}

fn chunk_loading(
	mut cmd: Commands,
	loaders: Query<&Pos, With<Loader>>,
	mut reg: ResMut<GeneratedRegistry>,
	mut updater: EventWriter<RenderUpdate>,
) {
	for loader in &loaders {
		for pos in chunks_surrounding(&loader.0, GEN_DISTANCE) {
			reg.load_chunk(&mut cmd, &mut updater, pos);
		}
	}
}

const GEN_DISTANCE: i32 = 1;
const CHUNK_WIDTH: i32 = 8;

fn chunks_surrounding(pos: &IVec2, radius: i32) -> Vec<IVec2> {
	let center_chunk = pos.div_euclid(IVec2::splat(CHUNK_WIDTH));
	(-radius..=radius)
		.flat_map(move |x| (-radius..=radius).map(move |y| center_chunk + IVec2::new(x, y)))
		.collect()
}

#[test]
fn test_euclid() {
	assert_eq!((5_i32).div_euclid(2), 2);
	assert_eq!((4_i32).div_euclid(2), 2);
	assert_eq!((3_i32).div_euclid(2), 1);
	assert_eq!((2_i32).div_euclid(2), 1);
	assert_eq!((1_i32).div_euclid(2), 0);
	assert_eq!((0_i32).div_euclid(2), 0);
	assert_eq!((-1_i32).div_euclid(2), -1);
	assert_eq!((-2_i32).div_euclid(2), -1);
	assert_eq!((-3_i32).div_euclid(2), -2);
	assert_eq!((-4_i32).div_euclid(2), -2);
	assert_eq!((-5_i32).div_euclid(2), -3);
}

#[test]
fn test_snap_mod() {
	assert_eq!((5_i32).rem_euclid(2), 1);
	assert_eq!((4_i32).rem_euclid(2), 0);
	assert_eq!((3_i32).rem_euclid(2), 1);
	assert_eq!((2_i32).rem_euclid(2), 0);
	assert_eq!((1_i32).rem_euclid(2), 1);
	assert_eq!((0_i32).rem_euclid(2), 0);
	assert_eq!((-1_i32).rem_euclid(2), 1);
	assert_eq!((-2_i32).rem_euclid(2), 0);
	assert_eq!((-3_i32).rem_euclid(2), 1);
	assert_eq!((-4_i32).rem_euclid(2), 0);
	assert_eq!((-5_i32).rem_euclid(2), 1);
}

#[derive(Debug, Resource, Default)]
pub struct GeneratedRegistry(HashSet<IVec2>);

impl GeneratedRegistry {
	pub fn load_chunk(&mut self, cmd: &mut Commands, rerender: &mut EventWriter<RenderUpdate>, chunk: IVec2) {
		if self.0.contains(&chunk) {
			return;
		}

		println!("Loading chunk indexed at position {chunk:?}.");
		let from = chunk * CHUNK_WIDTH;
		let to = from + IVec2::splat(CHUNK_WIDTH);
		for (pos, spawnable) in generate(from, to) {
			match spawnable {
				Spawnable::Block(display, mass) => spawn_block(cmd, rerender, pos, display, mass),
			}
		}
		self.0.insert(chunk);
	}
}

#[derive(Debug, Component)]
pub struct Loader;

#[derive(Debug, Component)]
pub struct Generated;

#[derive(Debug, Component)]
pub struct Chunk(IVec2);

enum Spawnable {
	Block((char, char), usize),
}

fn generate(from: IVec2, _to: IVec2) -> Vec<(IVec2, Spawnable)> {
	let structure = "
XX  XX
      
XX  XX
";
	map::str_to_struct(structure)
		.into_iter()
		.map(|(pos, display, mass)| (pos + from, Spawnable::Block(display, mass)))
		.collect()
}
