use bevy::{
	prelude::*,
	tasks::{TaskPool, TaskPoolBuilder},
};
use bevy_eventwork::{
	tcp::{NetworkSettings, TcpProvider},
	AppNetworkMessage, EventworkPlugin, EventworkRuntime, Network, NetworkData, NetworkEvent,
};

use crate::{
	common::{client_msg::ClientGreeting, server_msg::ServerGreeting},
	lib_server::player::{remove_player, spawn_player},
};

use super::{display::RenderUpdate, physics::Pos, player::Player};

pub struct NetPlugin;

impl Plugin for NetPlugin {
	fn build(&self, app: &mut App) {
		app.add_plugins(EventworkPlugin::<TcpProvider, TaskPool>::default())
			.insert_resource(EventworkRuntime(TaskPoolBuilder::new().num_threads(2).build()))
			.insert_resource(NetworkSettings::default())
			.add_systems(Update, handle_net_event)
			.add_systems(Startup, do_listen)
			.listen_for_message::<ClientGreeting, TcpProvider>()
			.add_systems(Update, respond_greet);
	}
}

fn handle_net_event(
	mut cmd: Commands,
	mut events: EventReader<NetworkEvent>,
	mut render_update: EventWriter<RenderUpdate>,
	players: Query<(Entity, &Player)>,
) {
	for event in events.read() {
		match event {
			NetworkEvent::Connected(conn_id) => {
				println!("New connection, {conn_id:?}.");
				spawn_player(&mut cmd, *conn_id, IVec2::ZERO);
				render_update.send(RenderUpdate(IVec2::ZERO));
			}
			NetworkEvent::Disconnected(conn_id) => {
				println!("Lost connection, {conn_id:?}.");
				remove_player(&mut cmd, *conn_id, &players);
			}
			NetworkEvent::Error(network_error) => println!("Network failure '{network_error}'."),
		}
	}
}

fn respond_greet(
	mut greetings: EventReader<NetworkData<ClientGreeting>>,
	players: Query<(&Pos, &Player)>,
	net: Res<Network<TcpProvider>>,
	mut render_update: EventWriter<RenderUpdate>,
) {
	for greet in greetings.read() {
		let source = *greet.source();
		net.send_message(source, ServerGreeting).ok();
		for (pos, player) in &players {
			if player.0 == source {
				render_update.send(RenderUpdate(pos.0));
			}
		}
	}
}

fn do_listen(
	mut net: ResMut<Network<TcpProvider>>,
	settings: Res<NetworkSettings>,
	task_pool: Res<EventworkRuntime<TaskPool>>,
) {
	net.listen(([0, 0, 0, 0], 9000).into(), &task_pool.0, &settings)
		.unwrap();
}
