use bevy::prelude::*;

use super::{display::RenderUpdate, player};

pub struct PhysicsPlugin;

impl Plugin for PhysicsPlugin {
	fn build(&self, app: &mut App) {
		app.add_systems(Update, update_physics_forces.after(player::handle_move))
			.add_systems(Update, update_physics_pos.after(update_physics_forces));
	}
}

#[derive(Debug, Component, Clone)]
pub struct Pos(pub IVec2);

#[derive(Debug, Component)]
pub struct Mass(pub usize);

#[derive(Debug, Component)]
pub struct Forces(pub IVec2);

impl Forces {
	pub fn zero() -> Self {
		Self::new(IVec2::ZERO)
	}

	pub fn new(forces: IVec2) -> Self {
		Self(forces)
	}

	pub fn add(&mut self, force: IVec2) {
		self.0 += force;
	}

	pub fn movement(&self, mass: usize) -> IVec2 {
		self.0 / mass as i32
	}

	pub fn movement_step(&self, mass: usize) -> IVec2 {
		self.movement(mass).clamp(IVec2::new(-1, -1), IVec2::new(1, 1))
	}
}

// 1. reduce force of adjascent objects.
pub fn update_physics_forces(mut forces: Query<(&mut Forces, &Pos, &Mass)>) {
	let mut combinations = forces.iter_combinations_mut();
	while let Some([left, right]) = combinations.fetch_next() {
		let (mut l_forces, l_pos, Mass(l_mass)) = left;
		let (mut r_forces, r_pos, Mass(r_mass)) = right;
		let l_step = l_forces.movement_step(*l_mass);
		let r_step = r_forces.movement_step(*r_mass);
		let (l_mass, r_mass) = (*l_mass as i32, *r_mass as i32);
		let same_step_x = r_step.x == l_step.x;
		let same_step_y = r_step.y == l_step.y;

		if l_pos.0.distance_squared(r_pos.0) > 2 {
			continue;
		}

		if !same_step_x {
			let l_will_overlap_x = l_pos.0 + IVec2::X * l_step.x == r_pos.0;
			let r_will_overlap_x = r_pos.0 + IVec2::X * r_step.x == l_pos.0;

			if l_will_overlap_x {
				r_forces.0.x += l_step.x * l_mass;
				l_forces.0.x -= l_step.x * r_mass.min(l_mass);
			}

			if r_will_overlap_x {
				l_forces.0.x += r_step.x * r_mass;
				r_forces.0.x -= r_step.x * l_mass.min(r_mass);
			}
		}

		if !same_step_y {
			let l_will_overlap_y = l_pos.0 + IVec2::Y * l_step.y == r_pos.0;
			let r_will_overlap_y = r_pos.0 + IVec2::Y * r_step.y == l_pos.0;

			if l_will_overlap_y {
				r_forces.0.y += l_step.y * l_mass;
				l_forces.0.y -= l_step.y * r_mass.min(l_mass);
			}

			if r_will_overlap_y {
				l_forces.0.y += r_step.y * r_mass;
				r_forces.0.y -= r_step.y * l_mass.min(r_mass);
			}
		}
	}
}

// 2. apply 1 step of forces.

pub fn update_physics_pos(mut bodies: Query<(&mut Forces, &mut Pos, &Mass)>, mut updater: EventWriter<RenderUpdate>) {
	for (mut forces, mut pos, Mass(mass)) in &mut bodies {
		let step = forces.movement_step(*mass);

		if step == IVec2::ZERO {
			continue;
		}

		updater.send(RenderUpdate(pos.0));
		pos.0 += step;
		forces.0 -= step * *mass as i32;
		updater.send(RenderUpdate(pos.0));
	}
}
