use crate::common::client_msg::Move;
use bevy::prelude::*;
use bevy_eventwork::{tcp::TcpProvider, AppNetworkMessage, ConnectionId, NetworkData};

use super::{
	display::{Display, RenderUpdate},
	generation::Loader,
	physics::{Forces, Mass, Pos},
};

pub struct PlayerPlugin;

impl Plugin for PlayerPlugin {
	fn build(&self, app: &mut App) {
		app.listen_for_message::<Move, TcpProvider>()
			.add_event::<RenderUpdate>()
			.add_systems(Update, handle_move);
	}
}

#[derive(Debug, Component)]
pub struct Player(pub ConnectionId);

#[derive(Debug, Component)]
pub struct Dir(pub IVec2);

fn player_display_for_dir(dir: IVec2) -> (char, char) {
	match dir {
		IVec2::Y => ('&', '/'),
		IVec2::NEG_X => ('_', '&'),
		IVec2::X => ('&', '_'),
		_ => ('&', '\\'),
	}
}

const PLAYER_MASS: usize = 10;

pub fn spawn_player(cmd: &mut Commands, conn_id: ConnectionId, pos: IVec2) {
	cmd.spawn((
		Player(conn_id),
		Loader,
		Pos(pos),
		Display(player_display_for_dir(IVec2::X)),
		Dir(IVec2::X),
		Forces::zero(),
		Mass(PLAYER_MASS),
	));
}

pub fn remove_player(cmd: &mut Commands, conn_id: ConnectionId, players: &Query<(Entity, &Player)>) {
	for (entity, player) in players {
		if player.0 == conn_id {
			cmd.entity(entity).despawn();
		}
	}
}

pub fn handle_move(
	mut moves: EventReader<NetworkData<Move>>,
	mut players: Query<(&Player, &mut Forces, &mut Dir, &mut Display)>,
) {
	for move_ in moves.read() {
		let conn_id = move_.source();
		for (Player(id), mut forces, mut dir, mut display) in &mut players {
			if conn_id == id {
				let mov_dir = move_.0.clamp(IVec2::splat(-1), IVec2::splat(1));
				forces.0 = mov_dir * PLAYER_MASS as i32;
				if !(mov_dir == IVec2::ZERO) {
					dir.0 = mov_dir;
					display.0 = player_display_for_dir(mov_dir);
				}
			}
		}
	}
}
